//-----------------------------------------------------------------------------
// FILE NAME:     conn.cpp
//
// AUTHOR:        Stefano Basagni
//
// LAST MODIFIED: April 2002
//
// DESCRIPTION:   Create random wireless topologies in 3D and verify if they
// are connected
// The executable can be run as
// > connect -n (number of nodes) -r (transmission range).0 -l (cube size)
//
//
//-----------------------------------------------------------------------------

#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <algo.h>

// STL containers for lists and arrays
#include <deque.h>
#include <vector.h>

// For inserting data from the command line (Linux)
#define _GNU_SOURCE
#include <getopt.h>

// Pseudo-random number generation
#include "rstream.h"


//----------------------------------------------------------------------
// PROTOTYPES
// Parsing of the parameters from the command line
void parsParam( int argc, char **argv, int &nodes, 
					 float &txp, int &l );

// Initialization of the nodes' coordinates
void initCoord( int n, int l, vector< float > &x, 
					 vector< float > &y, vector< float > &z );

// Builds the representation of the network (adj list)
void buildRep( int n, vector< float > x, vector< float > y, vector< float > z,
					float tpx, vector< deque<int> > &adjL );

// Test if the network is connected
bool isConnected( int n, vector< deque< int > > adjL );

// Neighbor test
bool isNeighbor( float x1, float y1, float z1, 
					  float x2, float y2, float z2, float txp );

// Printing out graph representations: Adj. List
// Used for testing
void printAList( int n, vector< deque< int > > l );


//---------------------------------------------------------------------
// MAIN
int main( int argc, char **argv ) {
  
  // MAIN VARIABLES of interest (possibly from the command line)
  
  // The number of nodes in the network
  int nn = 10;
  
  // Transmission power of nodes in grid units (m)
  float tx = 10.0; 
  
  // The size of the geographic area (a cube) (m)
  int L = 17;

  // The number of topologies generated for testing
  int topologies = 10;  
  
  // Read data of current run from the command line
  parsParam( argc, argv, nn, tx, L );
  

  for( int a = 0; a < topologies; a++ ) {

      vector< float > X( nn );            // x coordinates  
      vector< float > Y( nn );            // y coordinates
		vector< float > Z( nn );            // z coordinates
      vector< deque< int > > al( nn );   // Network adjacency list
    
		// The nodes coordinates are initialized
      initCoord( nn, L, X, Y, Z );
    
		// The network topology graph is built
      buildRep( nn, X, Y, Z, tx, al );

      if ( isConnected( nn, al ) ) 
		  cout << "Topology number: " << a << " is connected." << endl;
      else
		  cout << "Topology number: " << a << " is not connected." << endl;


  } // End for of the topologies

  return 0;

} // End main


//---------------------------------------------------------------------
// Parsing of the parameters of the current run from the command line
void parsParam( int argc, char **argv, int &nodes, 
		float &txp, int &l ) { 

  int c, index;
  static struct option long_options[] = {
    {"nodes", 1, 0, 'n'},
    {"txpw", 1, 0, 'r'},
    {"size", 1, 0, 'l'},
    {NULL, 0, 0, 0}
  };

  while (1) {
    c = getopt_long_only( argc, argv, "n:r:l:", long_options, &index );
    if ( c == EOF )
      break;
    switch (c) {
    case 'n':
      nodes = atol(optarg);
      break;
    case 'r':
      txp = atof(optarg);
      break;
    case 'l':
      l = atol(optarg);
      break;
    default:
      cout << "Ciao " << endl;
      exit(1);
      break;
    }
  }

} // End parsParam


//---------------------------------------------------------------------
// Initialization of the nodes' coordinates
void initCoord( int n, int l, vector< float > &x, vector< float > &y, 
					 vector< float > &z ) {
  
  // Random stream for the positions
  static Rstream pos( 1965 );
  
  for( int i = 0; i < n ; i++ ) {
    
    x[ i ] = pos.uniform( 0.0, (double)( l ) );
    y[ i ] = pos.uniform( 0.0, (double)( l ) );
	 z[ i ] = pos.uniform( 0.0, (double)( l ) );
  
  }

} // End initCoord


//---------------------------------------------------------------------
// Build the representation of the network (adj list)
void buildRep( int n, vector< float > x, vector< float > y, vector< float > z,
					float tpx, vector< deque< int > > &adjL ) {
  
  // Build up the adj. list
  for( int a = 0; a < n; a++ )
    for( int b =  a + 1; b < n; b++ ) 
      if ( isNeighbor( x[ a ], y[ a ], z[ a ], 
							  x[ b ], y[ b ], z[ b ], tpx ) )
		  {
			 (adjL[ a ]).push_front( b );
			 (adjL[ b ]).push_front( a );
		  }

} // End buildRep


//---------------------------------------------------------------------
// Depth-First Search (see Cormen+ book)
void DFS( int v, vector< deque< int > > al, vector< unsigned int > &c, 
          int &vis ) {

  c[ v ] = 1; // The node v is grayed

  for( unsigned int a = 0; a < al[ v ].size(); a++ )
    if ( c[ al[ v ][ a ] ] == 0 )
      DFS( al[ v ][ a ], al, c, vis );

  c[ v ] = 2; // The node v is blackened

  vis++;

}  // End DFS


//---------------------------------------------------------------------
// Test if the network is connected (based on DFS on the net adj list)
bool isConnected( int n, vector< deque< int > > adjL ) {

  vector< unsigned int > color( n, 0 ); // 0=white, 1=gray, 2=black

  int visited = 0;

  DFS( 0, adjL, color, visited );

  return ( visited == n );

} // End isConnected


//---------------------------------------------------------------------
// Distance (here Euclidean)
float distance( float x1, float y1, float z1, 
					 float x2, float y2, float z2 ) {

  return sqrt( pow( x1-x2, 2 ) + pow( y1-y2, 2 ) + pow( z1-z2, 2 ) );
  
} // End distance


//---------------------------------------------------------------------
// Neighbor test
bool isNeighbor( float x1, float y1, float z1, 
					  float x2, float y2, float z2, float txp ) {
  
  return ( distance( x1, y1, z1, x2, y2, z2 ) <= txp );
  
} // End isNeighbor


//---------------------------------------------------------------------
// Printing out graph representations: Adj. List
void printAList( int n, vector< deque< int > > l ) {

  for( int i = 0; i < n; i++ ) {
    cout << "Node " << i << " list:";
    for( unsigned int j = 0; j < l[ i ].size(); j++ )
      cout << " -> " << l[ i ][ j ];
    cout << endl; 
  }

} // End printAList

